# -*- coding: utf-8 -*-
# Copyright (C) 1998-2015 Mag. Christian Tanzer. All rights reserved
# Glasauergasse 32, A--1130 Wien, Austria. tanzer@swing.co.at
# ****************************************************************************
#
# This module is licensed under the terms of the BSD 3-Clause License
# <http://www.c-tanzer.at/license/bsd_3c.html>.
# ****************************************************************************
#
#++
# Name
#    TFL.EU_Currency
#
# Purpose
#    Provide classes for management of European Currencies
#
# Revision Dates
#    28-Dec-1998 (CT) Creation
#    31-Dec-1998 (CT) Rates entered
#    31-Jan-1999 (CT) `__cmp__', `__neg__', `__pos__', and `__abs__' added
#    27-Sep-1999 (CT) `command_spec' and `main' factored
#    29-Sep-1999 (CT) Use `Opt_L' for `-source' and `-target'
#     4-Jan-2000 (CT) `__float__' added
#     4-Jan-2000 (CT) `sep_1000' added
#    21-May-2001 (CT) `main': show evaluated argument (`b') if expression
#    29-Jul-2001 (CT) Inplace operators added
#    26-Dec-2001 (CT) Use attribute notation to access cmd-options
#    28-Dec-2001 (CT) Return statement added to inplace operators
#    29-Dec-2001 (CT) `as_source_s` added
#     1-Jan-2002 (CT) Argument `round_to_euro` added to `rounded` and its
#                     callers
#     1-Jan-2002 (CT) `_formatted` factored
#     5-Jan-2002 (CT) `EUR` alias added
#    12-Feb-2002 (CT) Argument `round_to_euro` renamed to `round`
#    12-Feb-2002 (CT) `rounded` corrected (must reset `cent` values < 50 for
#                     true values of `round`)
#    12-Feb-2002 (CT) `rounded_as_target` added
#    12-Feb-2002 (CT) `__nonzero__` added
#    13-Feb-2002 (CT) `rounded` corrected to handle negative numbers correctly
#    13-Feb-2002 (CT) `rounded_as_target` simplified
#    18-Jun-2003 (CT) `comma_dec_pat` added and used
#     5-Dec-2004 (CT) `to_euro_factor` changed from `1` to `1.0` to avoid
#                     `DeprecationWarning: classic int division`
#    11-Feb-2006 (CT) Moved into package `TFL`
#    17-Sep-2007 (CT) `EUC_Opt`, `EUC_Opt_SC`, and `EUC_Opt_TC` added
#    17-Sep-2007 (CT) Handling of `target_currency` simplified
#     8-May-2008 (CT) `EUC_Opt_TC` changed to use `__super`
#    30-Jun-2008 (CT) Parts of `EUC_Opt` factored to `Opt_L`
#     7-Mar-2009 (CT) Doctest added
#    11-Mar-2009 (CT) Refactored to derive from `TFL.Currency`
#    11-Mar-2009 (CT) `Euro` (based on decimal.Decimal but compatible to
#                     float-based `EU_Currency`) added
#     3-Jan-2010 (CT) `EUC_Source`, `EUC_Target`, and `_Command` added
#                     (all based on `TFL.CAO`)
#     7-Jan-2010 (CT) `_EUC_Currency_Arg_` added
#    16-Jun-2010 (CT) s/print/pyk.fprint/
#    17-Jun-2010 (CT) `__unicode__` introduced
#    29-Jan-2013 (CT) Adapt doctest to new option `Pdb_on_Exception`
#    16-Oct-2015 (CT) Add `__future__` imports
#    ««revision-date»»···
#--

from   __future__  import absolute_import
from   __future__  import division
from   __future__  import print_function
### importing `unicode_literals` from `__future__` wrecks 2.7 doctest

from   _TFL              import TFL

from   _TFL.pyk          import pyk
from   _TFL.Currency     import *
from   _TFL.Currency     import Currency, _Currency_

import _TFL.CAO

@pyk.adapt__str__
class _EU_Currency_ (_Currency_) :

    Table           = {}
    extension       = []

    ### if `target_currency' is not set, output is done in Euro
    target_currency = None

    to_euro_factor  = 1.0

    @classmethod
    def set_target_currency (cls, tc) :
        _EU_Currency_.target_currency = tc or Euro
    # end def set_target_currency

    def __str__ (self) :
        """Return `self.amount' as string representation of
           `self.target_currency'.
        """
        (amount, cent, target_currency) = self.as_target ()
        return u"%d%s%02d %s" % \
            ( amount
            , target_currency.decimal_sign
            , cent
            , target_currency.sloppy_name
            )
    # end def __str__

# end class _EU_Currency_

class Euro (_EU_Currency_, Currency) :
    """Model Euro currency using Decimal for representation."""

    def as_source_s (self, round = False) :
        return self.as_string (round)
    # end def as_source_s

    def as_string (self, round = False) :
        if EU_Currency.target_currency is Euro :
            return self.__super.as_string (round)
        else :
            return self._as_target_currency ().as_string (round)
    # end def as_string

    def as_string_s (self, round = 0) :
        if EU_Currency.target_currency is Euro :
            return self.__super.as_string_s (round)
        else :
            return self._as_target_currency ().as_string_s (round)
    # end def as_string_s

    def as_target (self, round = False, target_currency = None) :
        target_currency = target_currency or EU_Currency.target_currency
        if target_currency is Euro :
            a, c = self.rounded (round)
            return (a, c, target_currency)
        else :
            return self._as_target_currency (target_currency).as_target ()
    # end def as_target

    def rounded (self, amount, round = False) :
        if round :
            a, c = int (self), 0
        else :
            a, c = self.split ()
        return a, c
    # end def rounded

    def rounded_as_target (self) :
        if EU_Currency.target_currency is Euro :
            return self.quantize (self.U)
        else :
            return self._as_target_currency ().rounded_as_target ()
    # end def rounded_as_target

    def _massage_rhs (self, rhs) :
        if isinstance (rhs, EU_Currency) :
            rhs = rhs.amount
        rhs = self._massage_rhs_float (rhs)
        return rhs
    # end def _massage_rhs

    def _as_target_currency (self, target_currency = None) :
        ### we need to wrap the result in `target_currency` to get the right
        ### `decimal_sign`, `sep_1000`, ...
        target_currency = target_currency or EU_Currency.target_currency
        return target_currency (EU_Currency (float (self)))
    # end def _as_target_currency

# end class Euro

class EU_Currency (_EU_Currency_) :
    """Root class of hierarchy of Euro currencies"""

    C_Type          = property (lambda s : EU_Currency)

    def __init__ (self, amount = 0) :
        if isinstance (amount, (self.C_Type, Euro)) :
            self.amount = amount.amount
        else :
            self.amount = self.to_euro (amount)
    # end def __init__

    def to_euro (self, amount) :
        """Converts `amount' into Euro."""
        return float (amount) / self.to_euro_factor
    # end def to_euro

    def as_source_s (self, round = False) :
        """Return `self.amount` as string representation of `self.__class__`
           with 1000 separators.
        """
        (amount, cent, target_currency) = self.as_target \
            (round, self.__class__)
        result = self._formatted \
            (amount, cent, target_currency.decimal_sign, round)
        result = sep_1000_pat.sub \
            (r"\g<1>%s" % target_currency.sep_1000, result)
        return result
    # end def as_source_s

    def as_string (self, round = False) :
        """Return `self.amount' as string representation of
           `self.target_currency' (without currency name).
        """
        (amount, cent, target_currency) = self.as_target (round)
        return self._formatted \
            (amount, cent, target_currency.decimal_sign, round)
    # end def as_string

    def as_target (self, round = False, target_currency = None) :
        target_currency = target_currency or self.target_currency
        if target_currency is Euro :
            return Euro (str (self.amount)).as_target (round)
        else :
            target_currency = target_currency (0)
            amount          = self.amount * target_currency.to_euro_factor
            (amount, cent)  = target_currency.rounded (amount, round)
            return (amount, cent, target_currency)
    # end def as_target

    def normalized_amount (self, amount) :
        return amount
    # end def normalized_amount

    def rounded (self, amount, round = False) :
        """Return `amount' rounded to (euro, cent)."""
        euro = int (amount)
        cent = abs (int (((amount - euro) + 0.005) * 100))
        if cent == 100 :
            ### for some reason sometimes `cent == 100' results
            ### `amount' and `euro' differ by 1 in this case ???
            ### print "%f, %d, %f, %d" % (amount, euro, (amount - euro), cent)
            euro += 1
            cent  = 0
        if round :
            if cent >= 50 :
                if euro >= 0 :
                    euro += 1
                else :
                    euro -= 1
            cent  = 0
        return (euro, cent)
    # end def rounded

    def rounded_as_target (self) :
        (amount, cent, target_currency) = self.as_target (round = True)
        return target_currency.__class__ (amount)
    # end def rounded_as_target

    def _formatted (self, amount, cent, decimal_sign, round) :
        if round :
            return "%d"       % (amount, )
        else :
            return "%d%s%02d" % (amount, decimal_sign, cent)
    # end def _formatted

    def _massage_rhs (self, rhs) :
        if isinstance (rhs, Euro) :
            rhs = rhs.amount
        if isinstance (rhs, decimal.Decimal) :
            rhs = float (rhs)
        return rhs
    # end def _massage_rhs

    def __float__ (self) :
        return float (self.amount * self.target_currency.to_euro_factor)
    # end def __float__

    def __int__ (self) :
        return int (float (self))
    # end def __int__

# end class EU_Currency

def register (currency) :
    EU_Currency.Table [currency.name]                 = currency
    EU_Currency.Table [currency.name.lower ()]        = currency
    EU_Currency.Table [currency.name.upper ()]        = currency
    EU_Currency.Table [currency.sloppy_name]          = currency
    EU_Currency.Table [currency.sloppy_name.lower ()] = currency
    EU_Currency.Table [currency.sloppy_name.upper ()] = currency
    EU_Currency.extension.append (currency)
# end def register

class ATS (EU_Currency) :
    """Austrian currency ATS"""

    to_euro_factor = 13.7603
    name           = "ATS"
    sloppy_name    = u"öS"
    decimal_sign   = ","
    sep_1000       = "."

# end class ATS

class DEM (EU_Currency) :
    """German currency DEM"""

    to_euro_factor = 1.95583
    name           = "DEM"
    sloppy_name    = u"DM"

# end class DEM

class FRF (EU_Currency) :
    """French currency FRF"""

    to_euro_factor = 6.55957
    name           = "FRF"
    sloppy_name    = u"FF"

# end class FRF

class ITL (EU_Currency) :
    """Italian currency ITL"""

    to_euro_factor = 1936.27
    name           = "ITL"
    sloppy_name    = u"ITL"

# end class ITL

class BEF (EU_Currency) :
    """Belgian currency BEF"""

    to_euro_factor = 40.3399
    name           = "BEF"
    sloppy_name    = u"BF"

# end class BEF

class NLG (EU_Currency) :
    """Netherland's currency NLG"""

    to_euro_factor = 2.20371
    name           = "NLG"
    sloppy_name    = u"NLG"

# end class NLG

class ESP (EU_Currency) :
    """Spanish currency ESP"""

    to_euro_factor = 166.386
    name           = "ESP"
    sloppy_name    = u"ESP"

# end class ESP

class PTE (EU_Currency) :
    """Porugese currency PTE"""

    to_euro_factor = 200.482
    name           = "PTE"
    sloppy_name    = u"PTE"

# end class PTE

class FIM (EU_Currency) :
    """Finnish currency FIM"""

    to_euro_factor = 5.94573
    name           = "FIM"
    sloppy_name    = u"FIM"

# end class FIM

class IEP (EU_Currency) :
    """Irish currency IEP"""

    to_euro_factor = 0.787564
    name           = "IEP"
    sloppy_name    = u"IEP"

# end class IEP

class LUF (EU_Currency) :
    """Luxenburg's currency LUF"""

    to_euro_factor = 40.3399
    name           = "LUF"
    sloppy_name    = u"LUF"

# end class BEF

EUC = EU_Currency
EUC.set_target_currency (Euro)
EUR = Euro
for c in EUR, ATS, DEM, FRF, ITL, BEF, NLG, ESP, PTE, FIM, IEP, LUF :
    register (c)
EU_Currency.extension.sort (key = lambda c : c.name)

def currency (name) :
    return EU_Currency.Table [name]
# end def currency

class _EUC_Currency_Arg_ (TFL.CAO.Arg.Money) :
    """Argument or option with a currency value (epxressed in the currency
       specified by the option `-source_currency`).
    """

    _real_name = "EUC"

    def __init__ (self, ** kw) :
        self.sc_option_name = kw.pop ("sc_option_name", "source_currency")
        self.__super.__init__        (** kw)
    # end def __init__

    def cook (self, value, cao = None) :
        result = self.__super.cook (value)
        if result :
            SC = EUR
            if cao is not None :
                try :
                    SC = cao [self.sc_option_name]
                except KeyError :
                    pass
            result = SC (result)
        return result
    # end def cook

# end class _EUC_Currency_Arg_

class _EUC_Source_Arg_ (TFL.CAO.Arg.Key) :
    """Argument or option for source currency"""

    _real_name      = "EUC_Source"

    def __init__ (self, ** kw) :
        kw.setdefault ("name",        "source_currency")
        kw.setdefault ("default",     "EUR")
        kw.setdefault ("description", "Source currency")
        self.__super.__init__ (dct = EUC.Table, ** kw)
    # end def __init__

# end class _EUC_Source_Arg_

class _EUC_Target_Arg_ (TFL.CAO.Arg.Key) :
    """Argument or option for target currency"""

    _real_name = "EUC_Target"

    def __init__ (self, ** kw) :
        kw.setdefault ("name",        "target_currency")
        kw.setdefault ("default",     "EUR")
        kw.setdefault ("description", "Target currency")
        self.__super.__init__ (dct = EUC.Table, ** kw)
    # end def __init__

    def cook (self, value, cao = None) :
        result = self.__super.cook (value)
        if result :
            EUC.set_target_currency (result)
        return result
    # end def cook

# end class _EUC_Target_Arg_

def _main (cmd) :
    source = cmd.source_currency
    total  = source (0)
    for a in cmd.argv :
        c = source (a)
        pyk.fprint ("%s %s = %s" % (a, source.sloppy_name, c))
        total += c
    if total != 0 and len (cmd.argv) > 1 :
        pyk.fprint ("Total : %s" % total)
# end def _main

_Command = TFL.CAO.Cmd \
    ( handler     = _main
    , args        =
        ( TFL.CAO.Arg.Money
            ( name        = "amount"
            , description = "Amount to convert"
            , auto_split  = ";"
            )
        ,
        )
    , opts        =
        ( TFL.CAO.Opt.EUC_Source (default = "ATS")
        , TFL.CAO.Opt.EUC_Target ()
        )
    , description = "Convert between two Euro currencies"
    )

__doc__ = """
Provide classes for management of European Currencies compatible to
the Euro. The conversion factors are the ones published by the
European Union for the currencies that participated in the initial
introduction of the Euro.

    >>> for C in EU_Currency.extension :
    ...   pyk.fprint ("100 %s = %10s" % (C.name, C (100)))
    ...
    100 ATS =   7.27 EUR
    100 BEF =   2.48 EUR
    100 DEM =  51.13 EUR
    100 ESP =   0.60 EUR
    100 EUR = 100.00 EUR
    100 FIM =  16.82 EUR
    100 FRF =  15.24 EUR
    100 IEP = 126.97 EUR
    100 ITL =   0.05 EUR
    100 LUF =   2.48 EUR
    100 NLG =  45.38 EUR
    100 PTE =   0.50 EUR

    >>> EU_Currency.set_target_currency (ATS)
    >>> for C in EU_Currency.extension :
    ...   pyk.fprint ("100 %s = %10s" % (C.name, C (100)))
    ...
    100 ATS =  100,00 öS
    100 BEF =   34,11 öS
    100 DEM =  703,55 öS
    100 ESP =    8,27 öS
    100 EUR = 1376,03 öS
    100 FIM =  231,43 öS
    100 FRF =  209,77 öS
    100 IEP = 1747,20 öS
    100 ITL =    0,71 öS
    100 LUF =   34,11 öS
    100 NLG =  624,42 öS
    100 PTE =    6,86 öS

    >>> EU_Currency.set_target_currency (EU_Currency)
    >>> for C in EU_Currency.extension :
    ...   pyk.fprint ("100 %s + 100 ATS = %10s" % (C.name, C (100) + ATS (100)))
    ...
    100 ATS + 100 ATS =  14.53 EUR
    100 BEF + 100 ATS =   9.75 EUR
    100 DEM + 100 ATS =  58.40 EUR
    100 ESP + 100 ATS =   7.87 EUR
    100 EUR + 100 ATS = 107.27 EUR
    100 FIM + 100 ATS =  24.09 EUR
    100 FRF + 100 ATS =  22.51 EUR
    100 IEP + 100 ATS = 134.24 EUR
    100 ITL + 100 ATS =   7.32 EUR
    100 LUF + 100 ATS =   9.75 EUR
    100 NLG + 100 ATS =  52.65 EUR
    100 PTE + 100 ATS =   7.77 EUR

    >>> pyk.fprint (EUR (100) * 1.20)
    120.00 EUR

    >>> EUR (100) == 100.00
    True
    >>> ATS (100) == 7.27
    False
    >>> ATS (100) == 7.267283416785971
    True

    >>> print (TFL.ui_display (EUR (42)))
    42.00 EUR

This module extends TFL.CAO with three argument types:

- TFL.CAO.Arg.EUC_Source: argument or option to specify the source currency

- TFL.CAO.Arg.EUC_Target: argument or option to specify the target currency

- TFL.CAO.Arg.EUC: argument or option to specify a currency amount (using the
  source currency specified by the option `source_currency`).

These can be used like this:

    >>> cmd = TFL.CAO.Cmd (name = "Test", args = (TFL.CAO.Arg.EUC (name = "amount"), ), opts = (TFL.CAO.Opt.EUC_Source (), TFL.CAO.Opt.EUC_Target ()))
    >>> cao = cmd (["-help=vals"])
    Actual option and argument values of Test
        -Pdb_on_Exception   = False
        -help               = ['vals']
        -source_currency    = EUR
        -target_currency    = EUR
        amount              = None
            ()

    >>> cao = cmd (["-help=vals", "100"])
    Actual option and argument values of Test
        -Pdb_on_Exception   = False
        -help               = ['vals']
        -source_currency    = EUR
        -target_currency    = EUR
        amount              = 100
            100.00 EUR

    >>> cao = cmd (["-help=vals", "-source=ATS", "100.00"])
    Actual option and argument values of Test
        -Pdb_on_Exception   = False
        -help               = ['vals']
        -source_currency    = ATS
        -target_currency    = EUR
        amount              = 100.00
            7.27 EUR

    >>> pyk.fprint (cao.amount)
    7.27 EUR

"""

if __name__ == "__main__":
    _Command ()
### __END__ TFL.EU_Currency
