# -*- coding: utf-8 -*-
# Copyright (C) 2006-2016 Mag. Christian Tanzer. All rights reserved
# Glasauergasse 32, A--1130 Wien, Austria. tanzer@swing.co.at
# ****************************************************************************
#
# This module is licensed under the terms of the BSD 3-Clause License
# <http://www.c-tanzer.at/license/bsd_3c.html>.
# ****************************************************************************
#
#++
# Name
#    TFL.Recordifier
#
# Purpose
#    Provide classes supporting the conversion of formatted strings to records
#
# Revision Dates
#    17-Sep-2006 (CT) Creation
#    23-Dec-2010 (CT) Use `_print` for doctest (`%s` instead of `%r` for `v`)
#     9-Oct-2016 (CT) Move to Package_Namespace `TFL`
#     9-Oct-2016 (CT) Fix Python 3 compatibility
#    ««revision-date»»···
#--

from   __future__  import absolute_import
from   __future__  import division
from   __future__  import print_function
from   __future__  import unicode_literals

from   _TFL        import TFL
from   _TFL.pyk    import pyk

from   _TFL.Regexp import re

import _TFL.Caller
import _TFL.Record
import _TFL._Meta.Object

def _print (r) :
    print \
        ( "(%s)" % ", ".join \
            ( (   "%s = %s" % (k, v)
              for (k, v) in sorted (pyk.iteritems (r._kw))
              )
            )
        )
# end def _print

class _Recordifier_ (TFL.Meta.Object) :

    def __init__ (self, Result_Type) :
        self.Result_Type = Result_Type
    # end def __init__

    def __call__ (self, s) :
        conv   = self._converters
        result = self.Result_Type ()
        for k, v in self._field_iter (s) :
            setattr (result, k, conv [k] (v))
        return result
    # end def __call__

# end class _Recordifier_

class By_Regexp (_Recordifier_) :
    """Convert strings via regexp to records.

       >>> br = By_Regexp (
       ...   TFL.Regexp
       ...     (r"(?P<dt> (?P<y> \d{4})-(?P<m> \d{2})(?:-(?P<d> \d{2}))?)"
       ...      r" \s+ (?P<M> \d+) \s+ (?P<w> \d+\.\d*)", re.X)
       ...     , M = int, weight = float, y = int, m = int, d = int)
       >>> _print (br ("2006-06-01 6  96.4  1.20  93.5  98.1"))
       (M = 6, d = 1, dt = 2006-06-01, m = 6, w = 96.4, y = 2006)
       >>> _print (br ("2006-06 6  96.4  1.20  93.5  98.1"))
       (M = 6, dt = 2006-06, m = 6, w = 96.4, y = 2006)
    """

    field_pat = TFL.Regexp \
        ( r"\(\?P< (?P<name> [a-zA-Z_][a-zA-Z0-9_]*) >"
        , flags = re.VERBOSE
        )

    def __init__ (self, regexp, Result_Type = TFL.Record, ** converters) :
        self.__super.__init__ (Result_Type = Result_Type)
        self.regexp      = rex  = TFL.Regexp (regexp)
        self._converters = conv = {}
        for match in self.field_pat.search_iter (rex._pattern.pattern) :
            name = match.group ("name")
            conv [name] = \
                (  converters.get (name)
                or converters.get ("default_converter", str)
                )
    # end def __init__

    def _field_iter (self, s) :
        match  = self.regexp.search (s)
        if match :
            for k, v in pyk.iteritems (match.groupdict ()) :
                if v is not None :
                    yield k, v
        else :
            raise ValueError \
                ("`%s` doesn't match `%s`" % (s, self.regexp._pattern.pattern))
    # end def _field_iter

# end class By_Regexp

class By_Separator (_Recordifier_) :
    """Convert strings by splitting on whitespace into records.

       >>> bw = By_Separator (
       ...   "d", ("m", int), "avg", "err", "min", "max",
       ...   _default_converter = float, d = str)
       >>> _print (bw ("2006-06-01 6  96.4  1.20  93.5  98.1"))
       (avg = 96.4, d = 2006-06-01, err = 1.2, m = 6, max = 98.1, min = 93.5)
       >>> _print (bw ("2006-06-01 6  96.4  1.20  93.5"))
       (avg = 96.4, d = 2006-06-01, err = 1.2, m = 6, min = 93.5)
       >>> _print (bw ("2006-06-01 6  96.4  1.20  93.5  98.1 42"))
       (avg = 96.4, d = 2006-06-01, err = 1.2, m = 6, max = 98.1, min = 93.5)
    """

    _separator         = None
    _default_converter = str

    def __init__ (self, * fields, ** kw) :
        self.__super.__init__ \
            (Result_Type = kw.get ("Result_Type", TFL.Record))
        if "_separator" in kw :
            self._separator = kw ["_separator"]
        if "_default_converter" in kw :
            self._default_converter = kw ["_default_converter"]
        self._converters = conv = {}
        self._fields     = []
        add = self._fields.append
        for f in fields :
            if isinstance (f, pyk.string_types) :
                name    = f
                c       = kw.get (name, self._default_converter)
            else :
                name, c = f
            conv [name] = c
            add (name)
    # end def __init__

    def _field_iter (self, s) :
        for k, v in zip (self._fields, s.split ()) :
            yield k, v
    # end def _field_iter

# end class By_Separator

if __name__ == "__main__" :
    TFL._Export_Module ()
### __END__ TFL.Recordifier
