# -*- coding: utf-8 -*-
# Copyright (C) 2011-2015 Mag. Christian Tanzer All rights reserved
# Glasauergasse 32, A--1130 Wien, Austria. tanzer@swing.co.at
# ****************************************************************************
# This module is part of the package TFL.
#
# This module is licensed under the terms of the BSD 3-Clause License
# <http://www.c-tanzer.at/license/bsd_3c.html>.
# ****************************************************************************
#
#++
# Name
#    TFL.Trie
#
# Purpose
#    Trie (prefix tree)
#
# Revision Dates
#    15-Jan-2011 (CT) Creation
#    16-Jan-2011 (CT) `Node_P` and `Word_Trie_P` factored
#    16-Jan-2011 (CT) `matches_damerau` added (Damerau-Levenshtein)
#    17-Jan-2011 (CT) `completions` and `update` added
#    15-Apr-2012 (CT) Add `sorted` to guarantee determistic output
#                     for `PYTHONHASHSEED="random"`
#    23-May-2013 (CT) Use `TFL.Meta.BaM` for Python-3 compatibility
#    16-Oct-2015 (CT) Add `__future__` imports
#    ««revision-date»»···
#--

from   __future__  import absolute_import
from   __future__  import division
from   __future__  import print_function
### importing `unicode_literals` from `__future__` wrecks 2.7 doctest

from   _TFL        import TFL
from   _TFL.pyk    import pyk

import _TFL._Meta.M_Class
import _TFL._Meta.Object

import itertools

class Node (TFL.Meta.BaM (object, metaclass = TFL.Meta.M_Autosuper)) :
    """Node of a trie."""

    __slots__     = ("children", "value")

    kind          = "Node"
    parent        = None

    def __init__ (self, parent, value = None) :
        self.children = {}
        self.value    = value
    # end def __init__

    def pre_order (self) :
        """Generate node and all its children in pre-order."""
        def gen (key, node) :
            yield key, node
            for ck, cn in sorted (pyk.iteritems (node.children)) :
                for k, n in gen (ck, cn) :
                    yield k, n
        v = self.value
        return gen (v [-1] if v else "", self)
    # end def pre_order

    def values (self) :
        """Generate all values in node and its children."""
        for node in self :
            if node.value :
                yield node.value
    # end def values

    def visualized (self) :
        """Return string visualizing the node and its children."""
        children = self.children
        value    = self.value
        kind     = self.kind
        tail     = ""
        if children :
            tail = "\n  ".join \
                ( itertools.chain
                    ( * (   cn.visualized ().split ("\n")
                        for ck, cn in sorted (pyk.iteritems (children))
                        )
                    )
                )
        if value :
            if tail :
                fmt = "<Value %s\n  %s\n>"
            else :
                fmt = "<Value %s%s>"
            return fmt % (value, tail)
        elif children :
            return "<%s\n  %s\n>" % (kind, tail)
        else :
            return "<%s>" % (kind, )
    # end def visualized

    def __iter__ (self) :
        for key, node in self.pre_order () :
            yield node
    # end def __iter__

    def __repr__ (self) :
        children = self.children
        value    = self.value
        kind     = self.kind
        if value :
            if children :
                fmt = "<%s ...>"
            else :
                fmt = "<%s>"
            return fmt % (value, )
        else :
            return "<%s>" % (kind, )
    # end def __repr__

    if __debug__ :
        def __setattr__ (self, name, value) :
            if name == "value" :
                old = getattr (self, name, None)
                if not (value == old or value is None or old is None) :
                    raise AttributeError \
                        ("Cannot change value from %s to %s" % (old, value))
            object.__setattr__ (self, name, value)
        # end def __setattr__

# end class Node

class Node_P (Node) :
    """Node of a trie knowing its parent."""

    __slots__     = Node.__slots__ + ("parent", )

    kind          = property (lambda self : "Node" if self.parent else "Root")

    def __init__ (self, parent, value = None) :
        self.__super.__init__ (parent, value)
        self.parent = parent
    # end def __init__

# end class Node_P

class Word_Trie (TFL.Meta.Object) :
    """Trie (prefix tree) for words.

    >>> wt = Word_Trie (
    ...  ["ada", "adam", "beta", "cab", "cabby", "cat", "cats", "cathy", "cub"])
    >>> list (wt.values ())
    ['ada', 'adam', 'beta', 'cab', 'cabby', 'cat', 'cathy', 'cats', 'cub']
    >>> list (str (node) for key, node in  wt.pre_order ())
    ['<Node>', '<Node>', '<Node>', '<ada ...>', '<adam>', '<Node>', '<Node>',\
    '<Node>', '<beta>', '<Node>', '<Node>', '<cab ...>', '<Node>', '<cabby>',\
    '<cat ...>', '<Node>', '<cathy>', '<cats>', '<Node>', '<cub>']
    >>> list (repr (key) for key, node in  wt.pre_order ())
    ["''", "'a'", "'d'", "'a'", "'m'", "'b'", "'e'", "'t'", "'a'", "'c'",\
    "'a'", "'b'", "'b'", "'y'", "'t'", "'h'", "'y'", "'s'", "'u'", "'b'"]

    >>> wt.matches_levenshtein ("cat", 0)
    [('cat', 0)]
    >>> wt.matches_levenshtein ("cat", 1)
    [('cab', 1), ('cat', 0), ('cats', 1)]
    >>> wt.matches_levenshtein ("act", 1)
    []
    >>> wt.matches_levenshtein ("cat", 2)
    [('cab', 1), ('cat', 0), ('cathy', 2), ('cats', 1), ('cub', 2)]
    >>> wt.matches_levenshtein ("act", 2)
    [('ada', 2), ('cat', 2)]

    >>> wt.matches_damerau ("cat", 0)
    [('cat', 0)]
    >>> wt.matches_damerau ("cat", 1)
    [('cab', 1), ('cat', 0), ('cats', 1)]
    >>> wt.matches_damerau ("act", 1)
    [('cat', 1)]
    >>> wt.matches_damerau ("cat", 2)
    [('cab', 1), ('cat', 0), ('cathy', 2), ('cats', 1), ('cub', 2)]
    >>> wt.matches_damerau ("act", 2)
    [('ada', 2), ('cab', 2), ('cat', 1), ('cats', 2)]

    >>> wt.matches_levenshtein ("cit", 0)
    []
    >>> wt.matches_levenshtein ("cit", 1)
    [('cat', 1)]
    >>> wt.matches_levenshtein ("cit", 2)
    [('cab', 2), ('cat', 1), ('cats', 2), ('cub', 2)]

    >>> print (wt)
    Word_Trie (('ada', 'adam', 'beta', 'cab', 'cabby', 'cat', 'cathy',\
        'cats', 'cub'))
    >>> print (wt.visualized ())
    <Node
      <Node
        <Node
          <Value ada
            <Value adam>
          >
        >
      >
      <Node
        <Node
          <Node
            <Value beta>
          >
        >
      >
      <Node
        <Node
          <Value cab
            <Node
              <Value cabby>
            >
          >
          <Value cat
            <Node
              <Value cathy>
            >
            <Value cats>
          >
        >
        <Node
          <Value cub>
        >
      >
    >
    >>> print (wt.find ("ca").visualized ())
    <Node
      <Value cab
        <Node
          <Value cabby>
        >
      >
      <Value cat
        <Node
          <Value cathy>
        >
        <Value cats>
      >
    >

    >>> wt.completions("a")
    (('ada', 'adam'), False)
    >>> wt.completions("b")
    (('beta',), 'beta')
    >>> wt.completions("c")
    (('cab', 'cabby', 'cat', 'cathy', 'cats', 'cub'), False)
    >>> wt.completions("d")
    ((), False)
    >>> wt.completions("ca")
    (('cab', 'cabby', 'cat', 'cathy', 'cats'), False)
    >>> wt.completions("cat")
    (('cat', 'cathy', 'cats'), 'cat')
    >>> wt.completions ("cats")
    (('cats',), 'cats')
    """

    Node = Node

    def __init__ (self, * word_sets) :
        self.root = self.Node (None)
        if word_sets :
            self.update (* word_sets)
    # end def __init__

    def add (self, word) :
        """Add `word` to trie."""
        return self._add (word, self.root)
    # end def add

    def closest (self, word) :
        """Return node closest to `word`, indication if `word` is in trie,
           and key-length of result-node.
        """
        if word :
            result = self.root
            found  = True
            for i, c in enumerate (word) :
                try :
                    result = result.children [c]
                except KeyError :
                    found = False
                    break
            return result, found, i + bool (found)
    # end def closest

    def completions (self, prefix) :
        """Return all words in trie starting with `prefix` and unique
           completion for `prefix`, if any.
        """
        node, found, length = self.closest (prefix)
        if found :
            result = tuple (node.values ())
            unique = result and (len (result) == 1 or result [0] == prefix)
            return result, unique and result [0]
        return (), False
    # end def completions

    def discard (self, word) :
        """Remove `word` from trie; if `word` is not in trie, do nothing."""
        node, found, length = self.closest (word)
        if found :
            node.value = None
            for c in reversed (word) :
                parent = node.parent
                if node.children or not parent :
                    break
                del parent.children [c]
                node.parent = None
                node        = parent
        return found
    # end def discard

    def find (self, word) :
        """Return node containing `word`, if any."""
        node, found, length = self.closest (word)
        return found and node
    # end def find

    def longest_prefix (self, word) :
        """Return node closest to `word` and its key."""
        node, found, length = self.closest (word)
        return word [:length], node
    # end def longest_prefix

    @staticmethod
    def match_dict (matches) :
        from _TFL.multimap import mm_list
        result = mm_list ()
        for m, d in matches :
            result [d].append (m)
        return result
    # end def match_dict

    def match_iter_damerau (self, word, max_edits) :
        """Generate all matches with a Damerau-Levenshtein distance <= max_edits."""
        return self._match_iter (self._match_col_iter_d, word, max_edits)
    # end def match_iter_damerau

    def match_iter_levenshtein (self, word, max_edits) :
        """Generate all matches with a Levenshtein distance <= max_edits."""
        return self._match_iter (self._match_col_iter_l, word, max_edits)
    # end def match_iter_levenshtein

    def matches_damerau (self, word, max_edits, result_type = sorted) :
        """Return all matches with a Damerau-Levenshtein distance <= max_edits."""
        return result_type (self.match_iter_damerau (word, max_edits))
    # end def matches_damerau

    def matches_levenshtein (self, word, max_edits, result_type = sorted) :
        """Return all matches with a Levenshtein distance <= max_edits."""
        return result_type (self.match_iter_levenshtein (word, max_edits))
    # end def matches_levenshtein

    def pre_order (self) :
        """Generate all nodes in pre-order."""
        return self.root.pre_order ()
    # end def pre_order

    def remove (self, word) :
        """Remove `word` from trie."""
        if not self.discard (word) :
            raise KeyError (word)
    # end def remove

    def update (self, * word_sets) :
        """Add all words in `word_sets` to trie."""
        root = self.root
        for words in word_sets :
            for word in words :
                self._add (word, root)
    # end def update

    def values (self) :
        """Generate all values in trie."""
        return self.root.values ()
    # end def values

    def visualized (self) :
        """Return string visualizing the trie."""
        return self.root.visualized ()
    # end def visualized

    def _add (self, word, node) :
        for c in word :
            if c not in node.children :
                node.children [c] = self.Node (node)
            node = node.children [c]
        node.value = word
        return node
    # end def _add

    def _match_col_iter_d (self, word, char, node, row_c, row_1, row_2 = None, char_1 = None) :
        ### compute restricted Damerau-Levenshtein distance, aka,
        ### optimal string alignment
        ### http://en.wikipedia.org/wiki/Damerau-Levenshtein_distance
        for col, diff, edits in self._match_col_iter_l \
                (word, char, node, row_c, row_1, row_2, char_1) :
            if col > 1 and char_1 :
                if word [col - 1] == char_1 and word [col - 2] == char :
                    ### a transposition
                    edits = min (edits, row_2 [col - 2] + diff)
            yield col, diff, edits
    # end def _match_col_iter_d

    def _match_col_iter_l (self, word, char, node, row_c, row_1, row_2 = None, char_1 = None) :
        ### compute Levenshtein distance
        ### http://en.wikipedia.org/wiki/Levenshtein_distance
        for col in range (1, len (word) + 1) :
            diff = word [col - 1] != char
            yield col, diff, min \
                ( row_1 [col]     + 1    # a  deletion
                , row_c [col - 1] + 1    # an insertion
                , row_1 [col - 1] + diff # a  replacement
                )
    # end def _match_col_iter_l

    def _match_iter (self, col_iter, word, max_edits) :
        """Generate all words with a Levenshtein-distance <= max_edits to word."""
        ### http://en.wikipedia.org/wiki/Levenshtein_distance
        ### http://stevehanov.ca/blog/index.php?id=114
        row_1 = pyk.range (len (word) + 1)
        for char, node in pyk.iteritems (self.root.children) :
            for m in self._match_iter_inner \
                    (col_iter, word, max_edits, char, node, row_1) :
                yield m
    # end def _match_iter

    def _match_iter_inner (self, col_iter, word, max_edits, char, node, row_1, row_2 = None, char_1 = None) :
        row_c = [row_1 [0] + 1]
        for col, diff, edits in col_iter \
                (word, char, node, row_c, row_1, row_2, char_1) :
            ### append 1 by 1, because `_match_col_iter` needs `row_c [-1]`
            row_c.append (edits)
        ### row_c [-1] now contains the edit distance between
        ### `word` and `node.value`, if any
        if row_c [-1] <= max_edits and node.value is not None :
            yield (node.value, row_c [-1])
        if any (c <= max_edits for c in row_c) :
            char_1 = char
            for char, node in pyk.iteritems (node.children) :
                for m in self._match_iter_inner \
                        ( col_iter, word, max_edits
                        , char, node, row_c, row_1, char_1
                        ) :
                    yield m
    # end def _match_iter_inner

    def __iter__ (self) :
        return iter (self.root)
    # end def __iter__

    def __str__ (self) :
        return "%s ((%s))" % \
            ( self.__class__.__name__
            , ", ".join (repr (w) for w in self.values ())
            )
    # end def __str__

# end class Word_Trie

class Word_Trie_P (Word_Trie) :
    """Trie (prefix tree) for words which prunes empty elements on deletion.

    >>> wt = Word_Trie (
    ...  ["ada", "adam", "beta", "cab", "cabby", "cat", "cats", "cathy", "cub"])
    >>> wtp = Word_Trie_P (
    ...  ["ada", "adam", "beta", "cab", "cabby", "cat", "cats", "cathy", "cub"])

   >>> wt.discard ("cab")
    True
    >>> print (wt.find ("ca").visualized ())
    <Node
      <Node
        <Node
          <Value cabby>
        >
      >
      <Value cat
        <Node
          <Value cathy>
        >
        <Value cats>
      >
    >
    >>> wtp.discard ("cab")
    True
    >>> print (wtp.find ("ca").visualized ())
    <Node
      <Node
        <Node
          <Value cabby>
        >
      >
      <Value cat
        <Node
          <Value cathy>
        >
        <Value cats>
      >
    >

    >>> wt.discard ("cabby")
    True
    >>> print (wt.find ("ca").visualized ())
    <Node
      <Node
        <Node
          <Node>
        >
      >
      <Value cat
        <Node
          <Value cathy>
        >
        <Value cats>
      >
    >
    >>> wtp.discard ("cabby")
    True
    >>> print (wtp.find ("ca").visualized ())
    <Node
      <Value cat
        <Node
          <Value cathy>
        >
        <Value cats>
      >
    >

    >>> list (wt.values ())
    ['ada', 'adam', 'beta', 'cat', 'cathy', 'cats', 'cub']
    >>> list (wtp.values ())
    ['ada', 'adam', 'beta', 'cat', 'cathy', 'cats', 'cub']

    >>> list (str (node) for key, node in wt.pre_order ())
    ['<Node>', '<Node>', '<Node>', '<ada ...>', '<adam>', '<Node>', '<Node>',\
    '<Node>', '<beta>', '<Node>', '<Node>', '<Node>', '<Node>', '<Node>',\
    '<cat ...>', '<Node>', '<cathy>', '<cats>', '<Node>', '<cub>']
    >>> list (str (node) for key, node in wtp.pre_order ())
    ['<Root>', '<Node>', '<Node>', '<ada ...>', '<adam>', '<Node>', '<Node>',\
    '<Node>', '<beta>', '<Node>', '<Node>', '<cat ...>', '<Node>', '<cathy>',\
    '<cats>', '<Node>', '<cub>']

    >>> list (repr (key) for key, node in wt.pre_order ())
    ["''", "'a'", "'d'", "'a'", "'m'", "'b'", "'e'",\
    "'t'", "'a'", "'c'", "'a'", "'b'", "'b'", "'y'",\
    "'t'", "'h'", "'y'", "'s'", "'u'", "'b'"]
    >>> list (repr (key) for key, node in wtp.pre_order ())
    ["''", "'a'", "'d'", "'a'", "'m'", "'b'", "'e'",\
    "'t'", "'a'", "'c'", "'a'", "'t'", "'h'", "'y'",\
    "'s'", "'u'", "'b'"]

    """

    Node = Node_P

# end class Word_Trie_P

if __name__ != "__main__" :
    TFL._Export ("*")
### __END__ TFL.Trie
